package api

import (
	"bytes"
	"context"
	"errors"
	"math/big"
	"net"
	"witness/core"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	grpc_prometheus "github.com/grpc-ecosystem/go-grpc-prometheus"
	witnessv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/witness/v1"
	log "github.com/sirupsen/logrus"
	"google.golang.org/grpc"
)

type Server struct {
	witnessv1.UnimplementedWitnessServiceServer
	w *core.Witness
}

func (s *Server) WitnessStatus(ctx context.Context, req *witnessv1.WitnessStatusRequest) (resp *witnessv1.WitnessStatusResponse, err error) {
	return nil, nil
}

func (s *Server) PushProof(ctx context.Context, req *witnessv1.PushProofRequest) (resp *witnessv1.PushProofResponse, err error) {
	resp = &witnessv1.PushProofResponse{}
	proofs := req.GetProofs()
	if len(proofs) == 0 {
		return nil, errors.New("no proofs provided")
	}
	miner := common.HexToAddress(req.MinerAddress)
	reward := common.HexToAddress(req.RewardAddress)
	validatedProofs := make([]*witnessv1.ValidatedProof, 0, len(proofs))

	log.WithFields(log.Fields{"miner": miner, "reward": reward, "proofs": len(proofs)}).Debug("push proof")

	for _, proof := range proofs {
		if proof.Workload == 0 {
			continue
		}
		hashPayload := append(append([]byte(proof.TaskId), proof.ReqHash...), proof.RespHash...)
		containerPubKey, err := crypto.SigToPub(crypto.Keccak256(hashPayload), proof.ContainerSignature)
		if err != nil {
			log.WithError(err).Error("failed to verify container signature")
			return nil, err
		}
		containerSigner := crypto.PubkeyToAddress(*containerPubKey)
		verified := s.w.IsContainerAddress(containerSigner)
		if !verified {
			log.WithFields(log.Fields{"signer": containerSigner}).Error("invalid container signature")
			return nil, errors.New("invalid container signature")
		}

		minerHashPayload := append(append([]byte(proof.TaskId), proof.ReqHash...), proof.RespHash...)
		minerPubKey, err := crypto.SigToPub(crypto.Keccak256(minerHashPayload), proof.MinerSignature)
		if err != nil {
			log.WithError(err).Error("failed to verify miner signature")
			return nil, err
		}

		minerAddress := crypto.PubkeyToAddress(*minerPubKey)
		if minerAddress.Hex() != miner.Hex() {
			log.WithFields(log.Fields{"miner": miner, "signer": minerAddress}).Error("invalid miner address")
			return nil, errors.New("invalid miner address")
		}

		managerHashPayload := bytes.NewBuffer(hashPayload)
		managerHashPayload.Write(reward.Bytes())
		managerHashPayload.Write(miner.Bytes())
		managerHashPayload.Write(proof.ContainerSignature)
		managerHashPayload.Write(proof.MinerSignature)
		managerHashPayload.Write(big.NewInt(int64(proof.Workload)).Bytes())
		managerHashPayload.Write(big.NewInt(int64(proof.Timestamp)).Bytes())
		managerPubKey, err := crypto.SigToPub(crypto.Keccak256(managerHashPayload.Bytes()), proof.ManagerSignature)
		if err != nil {
			log.WithError(err).Error("failed to verify manager signature")
			return nil, err
		}

		managerSigner := crypto.PubkeyToAddress(*managerPubKey)
		verified = s.w.IsNMAddress(managerSigner)
		if !verified {
			log.WithFields(log.Fields{"signer": managerSigner}).Error("invalid manager signature")
			return nil, errors.New("invalid manager signature")
		}
		resp.Workload += proof.Workload
		validatedProofs = append(validatedProofs, &witnessv1.ValidatedProof{
			Workload: proof.Workload,
		})
	}
	if len(validatedProofs) > 0 {
		s.w.AddPendingProof(reward, validatedProofs)
	}
	return resp, nil
}

func (s *Server) GetMinerProof(ctx context.Context, req *witnessv1.GetMinerProofRequest) (*witnessv1.GetMinerProofResponse, error) {
	return nil, nil
}

func (s *Server) GetProof(ctx context.Context, req *witnessv1.GetProofRequest) (*witnessv1.GetProofResponse, error) {
	return nil, nil
}

func (s *Server) GetWithdrawProof(ctx context.Context, req *witnessv1.GetWithdrawProofRequest) (*witnessv1.GetWithdrawProofResponse, error) {
	return nil, nil
}

func StartGRPC(listenAddress string, w *core.Witness) {
	ln, err := net.Listen("tcp", listenAddress)
	if err != nil {
		log.WithError(err).Errorf("failed to listen on %s", listenAddress)
		return
	}

	log.WithField("listen", listenAddress).Info("start gRPC server")
	server := grpc.NewServer(grpc.MaxRecvMsgSize(1024*1024*1024), grpc.MaxSendMsgSize(1024*1024*1024))
	witnessv1.RegisterWitnessServiceServer(server, &Server{w: w})
	grpc_prometheus.Register(server)
	err = server.Serve(ln)
	if err != nil {
		log.WithError(err).Error("failed to serve")
		return
	}
	server.Stop()
}
